"""
Global Emission Spectra (GlobES)
================================

"""

from .globes import PyGCM
from .waccm import waccm_to_pygcm
from .exocam import exocam_to_pygcm
from . import structure
from .decoder import GCMdecoder



"""
A custom set of units to use in pypsg
"""

from astropy import units as u

ppm = u.def_unit(
    s=['ppm','ppmv'],
    represents=1e-6*u.dimensionless_unscaled,
    doc='part per million'
)
ppb = u.def_unit(
    s=['ppb','ppbv'],
    represents=1e-9*u.dimensionless_unscaled,
    doc='part per billion'
)
ppt = u.def_unit(
    s=['ppt','pptv'],
    represents=1e-12*u.dimensionless_unscaled,
    doc='part per trillion'
)
kbar = u.def_unit(
    s=['kbar'],
    represents=1e3*u.bar,
    doc='kilobar'
)
mbar = u.def_unit(
    s=['mbar'],
    represents=1e-3*u.bar,
    doc='milibar'
)
ubar = u.def_unit(
    s=['ubar'],
    represents=1e-6*u.bar,
    doc='microbar'
)

radiance_units = {
    'Wsrm2um': u.Unit('W sr-1 m-2 um-1'),
    'Wsrm2cm': u.Unit('W sr-1 m-2 cm-1'),
    'Wsrm2Hz': u.Unit('W sr-1 m-2 Hz-1'),
    'Jyarc': u.Unit('Jy arcsec-2'),
    'K' : u.K,
    'KRJ': u.K,
    'Wsrm2': u.Unit('W sr-1 m-2'),
    'Ra': u.astrophys.R,
    'Wsrum': u.Unit('W sr-1 um-1'),
    'Wsrcm': u.Unit('W sr-1 cm-1'),
    'Wsr': u.Unit('W sr-1'),
    'Wum': u.Unit('W um-1'),
    'Wcm': u.Unit('W cm-1'),
    'W': u.Unit('W'),
    'ph': u.ph/u.s,
    'pt': u.ph,
    'pm': u.ph,
    'Wm2': u.Unit('W m-2'),
    'erg': u.Unit('erg s-1 cm-2'),
    'Wm2um': u.Unit('W m-2 um-1'),
    'Wm2cm': u.Unit('W m-2 cm-1'),
    'Jy': u.Jy,
    'mJy': u.mJy,
    'rel': u.dimensionless_unscaled,
    'rif': u.dimensionless_unscaled,
    'V': u.mag
}

diameter = u.def_unit('diameter')
diffraction = u.def_unit('diffraction')
resolving_power = u.def_unit('resolving_power')

u.add_enabled_units([ppm,ppb,ppt,kbar,mbar,ubar,diameter,diffraction,resolving_power])


"""
This module is designed to read output files from exoCAM.

This code is based on PSG conversion scripts for exoCAM
written by Geronimo Villanueva

"""
import warnings
import requests
from typing import Tuple, Type
from netCDF4 import Dataset
from astropy import units as u
import numpy as np

from ...settings import psg_aerosol_size_unit, USER_DATA_PATH
from .. import structure
from ..globes import PyGCM
from .. import waccm as waccm

TIME_UNIT = u.day
ALBEDO_DEFAULT = 0.3
EMISSIVITY_DEFAULT = 1.0
DEFAULT_DESCRIPTION = 'exoCAM Atmosphere Model'
TEST_URL = 'https://archive.org/download/SimulatedPhaseDependentSpectraOfTerrestrialAquaplanetsInMdwarfSystems/t3000_s1550_p7.74511.cam.h0.avg_n68.nc?#mode=bytes'
TEST_PATH = USER_DATA_PATH / 'data' / 'exocam_test.nc'

REQUIRED_VARIABLES = [
    "hyam",
    "hybm",
    "P0",
    "PS",
    "T",
    "lat",
    "lon",
    "PS",
    "time",
    "time_bnds"
]
OPTIONAL_VARIABLES = [
    "TS",
    "ASDIR",
    "U",
    "V",
]

MOLEC_TRANSLATOR = {
    # "PSG" : "WACCM",
    'CO2': 'co2vmr',
    'CH4': 'ch4vmr',
    'N2O': 'n2ovmr',
    'H2O': 'Q'
}
"""
Translate molecule names to WACCM
"""
MOLEC_FILL_VALUE = 1e-30
"""
Fill in nans and infs
"""

AERO_TRANSLATOR = {
    # "PSG" : "WACCM",
    'Water': 'CLDLIQ',
    'WaterIce': 'CLDICE',
}
AERO_FILL_VALUE = 1e-30
"""
Fill in nans and infs
"""
AERO_SIZE_TRANSLATOR = {
    # "PSG" : "WACCM",
    'Water_size': 'REL',
    'WaterIce_size': 'REI',
}
AERO_SIZE_FILL_VALUE = 1e-6
"""
Fill in nans and infs
"""


class VariableAssumptionWarning(UserWarning):
    """
    A warning raised when a variable
    is not found in the netCDF file.
    """


def validate_variables(data: Dataset):
    """
    Check to make sure that the file
    contains all necessary variables.

    Parameters
    ----------
    data : netCDF4.Dataset
        The data to be checked
    """
    missing_vars = []
    for var in REQUIRED_VARIABLES:
        try:
            data.variables[var]
        except KeyError:
            missing_vars.append(var)
    if len(missing_vars) == 0:
        pass
    else:
        raise KeyError(
            f'Dataset is missing required variables: {",".join(missing_vars)}'
        )

    missing_vars = []
    for var in OPTIONAL_VARIABLES:
        try:
            data.variables[var]
        except KeyError:
            missing_vars.append(var)
    if len(missing_vars) == 0:
        pass
    else:
        warnings.warn(
            f'Dataset is missing optional variables: {",".join(missing_vars)}',
            VariableAssumptionWarning
        )


def get_time_index(data: Dataset, time: u.Quantity):
    """
    Get the index `itime` given a time quantity.

    Parameters
    ----------
    data : netCDF4.Dataset
        The dataset to use.
    time : astropy.units.Quantity
        The time Quantity.

    Returns
    -------
    itime : int
        The time index of `time`.
    """
    time_in_days = time.to_value(TIME_UNIT)
    time_left = data.variables['time_bnds'][:, 0]
    time_right = data.variables['time_bnds'][:, 1]
    itime = np.argwhere((time_in_days > time_left) &
                        (time_in_days <= time_right))[0][0]
    return itime


def get_shape(data: Dataset):
    """
    Get the shape of a Dataset

    Parameters
    ----------
    data : netCDF4.Dataset
        The dataset to use.

    Returns
    -------
    tuple
        The shape of `data`, (N_time,N_layers,N_lat,N_lon)
    """
    n_time = data.variables['T'].shape[0]
    n_layers = data.variables['T'].shape[1]
    n_lat = data.variables['T'].shape[2]
    n_lon = data.variables['T'].shape[3]
    return n_time, n_layers, n_lat, n_lon


def get_psurf(data: Dataset, itime: int) -> structure.SurfacePressure:
    """
    Get the surface pressure.

    Parameters
    ----------
    data : netCDF4.Dataset
        The dataset to use.
    itime : int
        The timestep to use.

    Returns
    -------
    psurf : structure.SurfacePressure
        The surface pressure
    """
    return waccm.waccm.get_psurf(data, itime)


def get_pressure(data: Dataset, itime: int) -> structure.Pressure:
    """
    Get the pressure.

    Parameters
    ----------
    data : netCDF4.Dataset
        The dataset to use.
    itime : int
        The timestep to use.

    Returns
    -------
    pressure : structure.Pressure
        The pressure.
    """
    return waccm.waccm.get_pressure(data, itime)
    


def get_temperature(data: Dataset, itime: int) -> structure.Temperature:
    """
    Get the temperature.

    Parameters
    ----------
    data : netCDF4.Dataset
        The dataset to use.
    itime : int
        The timestep to use.

    Returns
    -------
    temperature : structure.Temperature
        The temperature.
    """
    return waccm.waccm.get_temperature(data, itime)


def get_tsurf(data: Dataset, itime: int) -> structure.SurfaceTemperature:
    """
    Get the surface temperature.

    Parameters
    ----------
    data : netCDF4.Dataset
        The dataset to use.
    itime : int
        The timestep to use.

    Returns
    -------
    tsurf : structure.SurfaceTemperature
        The surface temperature.
    """
    return waccm.waccm.get_tsurf(data, itime)
    


def get_winds(data: Dataset, itime: int) -> Tuple[structure.Wind, structure.Wind]:
    """
    Get the winds.

    Parameters
    ----------
    data : netCDF4.Dataset
        The dataset to use.
    itime : int
        The timestep to use.

    Returns
    -------
    U : structure.Wind
        The wind speed in the U direction.
    V : structure.Wind
        The wind speed in the V direction.
    """
    return waccm.waccm.get_winds(data, itime)


def get_albedo(data: Dataset, itime: int) -> structure.Albedo:
    """
    Get the albedo.

    Parameters
    ----------
    data : netCDF4.Dataset
        The dataset to use.
    itime : int
        The timestep to use.

    Returns
    -------
    albedo : structure.Albedo
        The albedo.
    """
    return waccm.waccm.get_albedo(data, itime)

def get_emissivity(data: Dataset, itime: int) -> structure.Emissivity:
    """
    Get the emissivity.

    Parameters
    ----------
    data : netCDF4.Dataset
        The dataset to use.
    itime : int
        The timestep to use.

    Returns
    -------
    emissivity : structure.Emissivity
        The Emissivity.
    """
    return waccm.waccm.get_emissivity(data, itime)


def _generic_getter(data: Dataset, itime: int, name: str, translator: dict, fill_value: float, unit: u.Unit, cls: Type, mean_molec_mass:float=None):
    """
    Generic getter for a variable.
    """
    dat = np.flip(
        np.array(data.variables[translator.get(name, name)][itime, :, :, :]), axis=0)
    dat = np.where((dat > 0) & (np.isfinite(dat)), dat, fill_value) * unit
    if name == 'H2O':
        if mean_molec_mass is None:
            raise ValueError('Mean molecular mass must be specified for H2O.')
        dat = dat/ (1 - dat)
        dat = dat * (mean_molec_mass/18.0)
    dat = np.swapaxes(dat, 1, 2)
    return cls(name, dat)


def get_molecule(data: Dataset, itime: int, name: str, mean_molecular_mass: float=None) -> structure.Molecule:
    """
    Get the abundance of a molecule.

    Parameters
    ----------
    data : netCDF4.Dataset
        The dataset to use.
    itime : int
        The timestep to use.
    name : str
        The variable name of the molecule.
    mean_molecular_mass : float, optional
        The mean molecular mass of the atmosphere. This is required to extract
        the water abundance from the specific humidity variable.

    Returns
    -------
    molec : structure.Molecule
        The concentration of the molecule.
    """
    return _generic_getter(data, itime, name, MOLEC_TRANSLATOR, MOLEC_FILL_VALUE, u.dimensionless_unscaled, structure.Molecule, mean_molecular_mass)


def get_aerosol(data: Dataset, itime: int, name: str) -> structure.Aerosol:
    """
    Get the abundance of an aerosol.

    Parameters
    ----------
    data : netCDF4.Dataset
        The dataset to use.
    itime : int
        The timestep to use.
    name : str
        The variable name of the aerosol.

    Returns
    -------
    aero : structure.Aerosol
        The concentration of the aerosol.
    """
    return _generic_getter(data, itime, name, AERO_TRANSLATOR, MOLEC_FILL_VALUE, u.dimensionless_unscaled, structure.Aerosol)


def get_aerosol_size(data: Dataset, itime: int, name: str) -> structure.AerosolSize:
    """
    Get the size of an aerosol.

    Parameters
    ----------
    data : netCDF4.Dataset
        The dataset to use.
    itime : int
        The timestep to use.
    name : str
        The variable name of the aerosol.

    Returns
    -------
    aero : structure.Aerosol
        The concentration of the aerosol.
    """
    return _generic_getter(data, itime, f'{name}_size', AERO_SIZE_TRANSLATOR, AERO_SIZE_FILL_VALUE, psg_aerosol_size_unit, structure.AerosolSize)


def get_molecule_suite(data: Dataset, itime: int, names: list, background: str = None, mean_molecular_mass: float=None) -> Tuple[structure.Molecule]:
    """
    Get the abundance of a suite of molecules.

    Parameters
    ----------
    data : netCDF4.Dataset
        The dataset to use.
    itime : int
        The timestep to use.
    names : list of str
        The variable names of the molecules.
    background : str, default=None
        The variable name of a background gas to include.

    Returns
    -------
    molec : tuple of structure.Molecule
        The molecules in the GCM
    """
    molecs = tuple(get_molecule(data, itime, name, mean_molecular_mass) for name in names)
    if background is not None:
        if background in names:
            raise ValueError(
                'Do not know how to handle specifying a background'
                'gas that is already in our dataset.'
            )
        else:
            _, n_layer, n_lat, n_lon = get_shape(data)
            background_abn = np.ones(
                shape=(n_layer, n_lon, n_lat))*u.dimensionless_unscaled
            for molec in molecs:
                background_abn -= molec.dat
            if np.any(background_abn < 0):
                raise ValueError('Cannot have negative abundance.')
            molecs += (structure.Molecule(background, background_abn),)
    return molecs

def to_pygcm(
    data:Dataset,
    itime:int,
    molecules:list,
    aerosols:list,
    background=None,
    lon_start:float=-180.,
    lat_start:float=-90.,
    desc:str=DEFAULT_DESCRIPTION,
    mean_molecular_mass:float=None
)->PyGCM:
    """
    Covert a WACCM dataset to a Planet object.
    
    Parameters
    ----------
    data : netCDF4.Dataset
        The GCM dataset.
    itime : int
        The time index.
    molecules : list
        The variable names of the molecules.
    aerosols : list
        The variable names of the aerosols.
    background : str, optional
        The optional background gas to assume.
    lon_start : float, optional
        The starting longitude of the GCM. Defaults to -180.
    lat_start : float, optional
        The starting latitude of the GCM. Defaults to -90.
    desc : str, optional
        The description of the GCM.'
    mean_molecular_mass : float, optional
        The mean molecular mass of the atmosphere. Defaults to None.
    """
    molecules:tuple = tuple() if molecules is None else get_molecule_suite(data,itime,molecules,background,mean_molecular_mass)
    
    _aerosols:tuple = tuple() if aerosols is None else tuple(get_aerosol(data,itime,name) for name in aerosols)
    aerosol_sizes:tuple = tuple() if aerosols is None else tuple(get_aerosol_size(data,itime,name) for name in aerosols)
    
    wind_u, wind_v = get_winds(data,itime)
    
    return PyGCM(
        get_pressure(data,itime),
        get_temperature(data,itime),
        *(molecules + _aerosols + aerosol_sizes),
        wind_u=wind_u,
        wind_v=wind_v,
        tsurf=get_tsurf(data,itime),
        psurf=get_psurf(data,itime),
        albedo=get_albedo(data,itime),
        emissivity=get_emissivity(data,itime),
        lon_start=lon_start,
        lat_start=lat_start,
        desc=desc
    )

def download_test_data(rewrite=False):
    """
    Download the WACCM test data.
    """
    
    
    TEST_PATH.parent.mkdir(exist_ok=True)
    if TEST_PATH.exists() and not rewrite:
        return TEST_PATH
    else:
        TEST_PATH.unlink(missing_ok=True)
        with requests.get(TEST_URL,stream=True,timeout=20) as req:
            req.raise_for_status()
            with TEST_PATH.open('wb') as f:
                for chunk in req.iter_content(chunk_size=8192):
                    f.write(chunk)
        return TEST_PATH


"""
Test the user interface
"""
from pathlib import Path
import pytest
import requests
from astropy import units as u
import time

from pypsg.cfg import PyConfig, BinConfig, models
from pypsg import settings
from pypsg.globes import globes
from pypsg.exceptions import GlobESError, PSGMultiError
from pypsg.settings import INTERNAL_PSG_URL, PSG_URL
from pypsg.docker import start_psg, stop_psg, is_psg_installed
from pypsg import APICall, PSGResponse, PyRad, PyLyr

TR1e_PATH = Path(__file__).parent / 'test_cfg' / 'data' / 'TR1e_mirecle.cfg'

@pytest.fixture
def keep_psg_settings():
    """
    Make sure my PSG settings don't change after running tests.
    Otherwise I could not test different settings.
    """
    orginal_url = settings.get_setting('url')
    yield None
    settings.save_settings(url=orginal_url)
    
@pytest.fixture
def temp_file():
    """
    Get a temporary file name and make sure it gets deleted.
    """
    file_path = Path(__file__).parent / 'temp.txt'
    yield file_path
    file_path.unlink(missing_ok=True)

@pytest.mark.slow
def test_apicall(keep_psg_settings):
    """
    Test user interaction with the APICall class
    
    Parameters
    ----------
    cfg : Config or BinConfig
        A configuration object. We won't worry about its contents for this test, but it can be a `BinConfig` or a `PyConfig`.
    output_type : str or None
        The type of output to ask for. This could be a specific file type, set, update, all, or None
    app : str or None
        The app to use. Either None of 'globes'. We will just test None for now.
    url : str or None
        The URL to send the request to. This can be the external URL or the internal URL. It also can be None.
    
    Public Attributes
    -----------------
    None
    
    Public Methods
    --------------
    __call__(self)
        Calls the PSG API and returns a `PSGResponse` object.
    
    Other Expectations
    ------------------
    * If PSG is not running and the internal URL is specified, an exception will be raised.
    * If globes is used but there is not GCM an exception will be raised.
        - This seems to not be the case when output is `all` and the docker version is used.
    
    """
    
    cfgs = [PyConfig.from_file(TR1e_PATH), BinConfig.from_file(TR1e_PATH)]
    output_types = [None,'all','cfg', 'rad', 'lyr', 'noi', 'upd', 'set']
    apps = [None]
    urls = [None, INTERNAL_PSG_URL, PSG_URL] if is_psg_installed() else [None, PSG_URL]
    url_settings = [INTERNAL_PSG_URL, PSG_URL] if is_psg_installed() else [PSG_URL]
    is_psg_running = [False, True] if is_psg_installed() else [False]    
    
    def check(
        psg_running,
        cfg,
        output_type,
        app,
        url,
        url_setting
    ):
        api = APICall(
            cfg=cfg,
            output_type=output_type,
            app=app,
            url=url
            )
        resolved_url = url if url is not None else url_setting
        expect_globes_err = app == 'globes'
        expect_globes_err = expect_globes_err and output_type not in ['set','upd']
        
        if not psg_running and resolved_url == INTERNAL_PSG_URL:
            with pytest.raises(requests.ConnectionError):
                time.sleep(0.1)
                _ = api()
        elif expect_globes_err:
            with pytest.raises((GlobESError,PSGMultiError)):
                _ = api()
            APICall(cfg=cfg, output_type='set', app=None, url=url)() # This should clean up
        else:
            time.sleep(0.1)
            response = api()
            assert isinstance(response, PSGResponse)
            match output_type:
                case 'cfg':
                    assert isinstance(response.cfg, PyConfig)
                    assert response.rad is None
                    assert response.lyr is None
                    assert response.noi is None
                    assert response.cfg.target.object.value == 'Exoplanet'
                case 'rad':
                    assert isinstance(response.rad, PyRad)
                    assert response.cfg is None
                    assert response.lyr is None
                    assert response.noi is None
                case 'lyr':
                    assert isinstance(response.lyr, PyLyr)
                    assert response.cfg is None
                    assert response.rad is None
                    assert response.noi is None
                case 'noi':
                    assert isinstance(response.noi, PyRad)
                    assert response.cfg is None
                    assert response.rad is None
                    assert response.lyr is None
                case 'all':
                    assert isinstance(response.cfg, PyConfig)
                    assert isinstance(response.rad, PyRad)
                    assert isinstance(response.lyr, PyLyr)
                    assert isinstance(response.noi, PyRad)
                case 'upd':
                    assert response.cfg is None
                    assert response.rad is None
                    assert response.lyr is None
                    assert response.noi is None
                case 'set':
                    assert response.cfg is None
                    assert response.rad is None
                    assert response.lyr is None
                    assert response.noi is None
                case None:
                    assert response.cfg is None
                    assert isinstance(response.rad, PyRad)
                    assert response.lyr is None
                    assert response.noi is None
            # APICall(cfg=cfg, output_type='set', app=None, url=url)() # This should clean up
    
    # Test all combinations
    n_tested = 0
    for psg_running in is_psg_running:
        if psg_running:
            start_psg(strict=True)
        else:
            stop_psg(strict=False)
        for url_setting in url_settings:
            settings.save_settings(url=url_setting)
            for cfg in cfgs:
                for output_type in output_types:
                    for app in apps:
                        for url in urls:
                            n_tested += 1
                            try:
                                check(
                                    psg_running,
                                    cfg,
                                    output_type,
                                    app,
                                    url,
                                    url_setting
                                )
                            except Exception as e:
                                msg = f'Failed for\npsg_running={psg_running}\ncfg={cfg}\noutput_type={output_type}\napp={app}\nurl={url}\nurl_setting={url_setting}\nn_tested={n_tested}'
                                raise Exception(msg) from e
                            

class TestPyConfig:
    """
    Test user interaction with the PyConfig class
    """
    
    def test_init(self):
        """
        Test the initialization of the PyConfig class
        
        Parameters
        ----------
        target : models.Target
            Target object
        geometry : models.Geometry
            Geometry object
        atmosphere : models.Atmosphere
            Atmosphere object
        surface : models.Surface
            Surface object
        generator : models.Generator
            Generator object
        telescope : models.Telescope
            Telescope object
        noise : models.Noise
            Noise object
        gcm : globes.GCM
            GCM object
        
        Expectations
        ------------
        Each argument will either be kept as is or (in the case where None is passed),
        replaced with the default (empty) version of the Model.
        
        The exception is for `gcm`, which is kept as is always.
        """
        # Case where everything is None
        cfg = PyConfig()
        assert isinstance(cfg.target, models.Target), f'Target is {type(cfg.target)}'
        assert cfg.target.content == b'', f'Target content is {cfg.target.content}'
        assert isinstance(cfg.geometry, models.Geometry), f'Geometry is {type(cfg.geometry)}'
        assert cfg.geometry.content == b'', f'Geometry content is {cfg.geometry.content}'
        assert isinstance(cfg.atmosphere, models.Atmosphere), f'Atmosphere is {type(cfg.atmosphere)}'
        assert cfg.atmosphere.content == b'', f'Atmosphere content is {cfg.atmosphere.content}'
        assert isinstance(cfg.surface, models.Surface), f'Surface is {type(cfg.surface)}'
        assert cfg.surface.content == b'', f'Surface content is {cfg.surface.content}'
        assert isinstance(cfg.generator, models.Generator), f'Generator is {type(cfg.generator)}'
        assert cfg.generator.content == b'', f'Generator content is {cfg.generator.content}'
        assert isinstance(cfg.telescope, models.Telescope), f'Telescope is {type(cfg.telescope)}'
        assert cfg.telescope.content == b'', f'Telescope content is {cfg.telescope.content}'
        assert isinstance(cfg.noise, models.Noise), f'Noise is {type(cfg.noise)}'
        assert cfg.noise.content == b'', f'Noise content is {cfg.noise.content}'
        assert cfg.gcm is None, f'GCM is {type(cfg.gcm)}'
        
        # Case with Target
        cfg = PyConfig(target=models.Target(object='Exoplanet',name='earth',))
        assert isinstance(cfg.target, models.Target), f'Target is {type(cfg.target)}'
        assert cfg.target.object.value == 'Exoplanet'
        assert cfg.target.name.value == 'earth'
        assert cfg.target.date.value is None
        
        # Case with Geometry
        cfg = PyConfig(geometry=models.Observatory(stellar_type='M'))
        assert isinstance(cfg.geometry, models.Geometry), f'Geometry is {type(cfg.geometry)}'
        assert cfg.geometry.stellar_type.value == 'M'
        assert cfg.geometry.geometry.value == 'Observatory'
        assert cfg.geometry.ref.value is None
        
        # Case with Atmosphere
        cfg = PyConfig(atmosphere=models.EquilibriumAtmosphere(pressure=1*u.bar))
        assert isinstance(cfg.atmosphere, models.Atmosphere), f'Atmosphere is {type(cfg.atmosphere)}'
        assert cfg.atmosphere.structure.value == 'Equilibrium'
        assert cfg.atmosphere.pressure.value == 1*u.bar
        assert cfg.atmosphere.temperature.value is None
        
        # Case with Surface
        cfg = PyConfig(surface=models.Surface(temperature=100*u.K))
        assert isinstance(cfg.surface, models.Surface), f'Surface is {type(cfg.surface)}'
        assert cfg.surface.temperature.value == 100*u.K
        assert cfg.surface.albedo.value is None
        assert cfg.surface.emissivity.value is None
        
        # Case with Generator
        cfg = PyConfig(generator=models.Generator(gas_model=True))
        assert isinstance(cfg.generator, models.Generator), f'Generator is {type(cfg.generator)}'
        assert cfg.generator.gas_model.value is True
        assert cfg.generator.continuum_model.value is None
        
        # Case with Telescope
        cfg = PyConfig(telescope=models.SingleTelescope(apperture=10*u.m))
        assert isinstance(cfg.telescope, models.Telescope), f'Telescope is {type(cfg.telescope)}'
        assert cfg.telescope.telescope.value == 'SINGLE'
        assert cfg.telescope.apperture.value == 10*u.m
        assert cfg.telescope.zodi.value is None
        
        # Case with Noise
        cfg = PyConfig(noise=models.CCD(temperature=35*u.K))
        assert isinstance(cfg.noise, models.Noise), f'Noise is {type(cfg.noise)}'
        assert cfg.noise.noise_type.value == 'CCD'
        assert cfg.noise.temperature.value == 35*u.K
        assert cfg.noise.read_noise.value is None
        
        # Case with GCM
        cfg = PyConfig(gcm=globes.GCM('header',[1,2,3]))
        assert isinstance(cfg.gcm, globes.GCM), f'GCM is {type(cfg.gcm)}'
    
    def test_from_dict(self):
        """
        Test initalization from a dictionary.
        
        Parameters
        ----------
        d : dict
            A dictionary representation of a PSG config file.
            Under the hood, the actual dictionary is parsed
            by each field individually, but the user does not
            know or care about that.
        
        Considerations
        --------------
        * The dictionary has a single level of nesting.
        * Additional keys are ignored.
        * The dictionary is not validated.
        * Some fields use more than one key.
        * Any key that is lowercase will raise an error.
        """
        
        # Nominal case
        d = {'OBJECT-NAME': 'Earth', 'GEOMETRY': 'Observatory', 'GEOMETRY-OBS-ALTITUDE': 1.3, 'GEOMETRY-ALTITUDE-UNIT': 'pc'}
        cfg = PyConfig.from_dict(d)
        assert isinstance(cfg.target, models.Target), f'Target is {type(cfg.target)}'
        assert cfg.target.name.value == 'Earth'
        assert isinstance(cfg.geometry, models.Geometry), f'Geometry is {type(cfg.geometry)}'
        assert cfg.geometry.geometry.value == 'Observatory'
        assert cfg.geometry.observer_altitude.value == 1.3*u.pc
        
        # Extra key
        d = {'OBJECT-NAME': 'Earth', 'EXTRA': 'foo'}
        cfg = PyConfig.from_dict(d)
        assert isinstance(cfg.target, models.Target), f'Target is {type(cfg.target)}'
        assert cfg.target.name.value == 'Earth'
        
        # Missing key
        d = {'OBJECT-NAME': 'Earth', 'GEOMETRY': 'Observatory', 'GEOMETRY-OBS-ALTITUDE': 1.3}
        cfg = PyConfig.from_dict(d)
        assert isinstance(cfg.target, models.Target), f'Target is {type(cfg.target)}'
        assert cfg.target.name.value == 'Earth'
        assert isinstance(cfg.geometry, models.Geometry), f'Geometry is {type(cfg.geometry)}'
        assert cfg.geometry.geometry.value == 'Observatory'
        assert cfg.geometry.observer_altitude.value is None
        
        # Lowercase key
        d = {'object-name': 'Earth'}
        with pytest.raises(ValueError):
            _ = PyConfig.from_dict(d)
    
    def test_from_binaryconfig(self):
        """
        Test initalization from a binary config object.
        
        Parameters
        ----------
        bcfg : BinConfig
            A binary config object.
        """
        b = b'<OBJECT-NAME>Earth'
        cfg = PyConfig.from_binaryconfig(BinConfig(b))
        assert isinstance(cfg.target, models.Target), f'Target is {type(cfg.target)}'
        assert cfg.target.name.value == 'Earth'
    
    def test_from_bytes(self):
        """
        Test initalization from a bytes object.
        
        Parameters
        ----------
        b : bytes
            A bytes object.
        
        Considerations
        --------------
        * The bytes object is not validated.
        * The bytes object is first read as a BinConfig, so it should
          be responsible for handling any errors.
        """
        b = b'<OBJECT-NAME>Earth'
        cfg = PyConfig.from_bytes(b)
        assert isinstance(cfg.target, models.Target), f'Target is {type(cfg.target)}'
        assert cfg.target.name.value == 'Earth'
    
    def test_from_file(self):
        """
        Test initalization from a file path.
        
        Parameters
        ----------
        path : Path | str
            A file path.
        """
        # pathlib.Path
        path = Path(__file__).parent / 'test_cfg' / 'data' / 'object_gj1214b.cfg'
        cfg = PyConfig.from_file(path)
        assert isinstance(cfg.target, models.Target), f'Target is {type(cfg.target)}'
        assert cfg.target.name.value == 'GJ 1214b'
        
        # str
        path = str(Path(__file__).parent / 'test_cfg' / 'data' / 'object_gj1214b.cfg')
        cfg = PyConfig.from_file(path)
        assert isinstance(cfg.target, models.Target), f'Target is {type(cfg.target)}'
        assert cfg.target.name.value == 'GJ 1214b'
    
    def test_content(self):
        """
        Test the `content` property.
        """
        
        cfg = PyConfig(target=models.Target(name='Earth'))
        assert cfg.content == b'<OBJECT-NAME>Earth'
        
    def test_tofile(self, temp_file:Path):
        """
        Test writing the config to a file.
        """
        
        cfg = PyConfig(target=models.Target(name='Earth'))
        cfg.to_file(temp_file)
        assert temp_file.read_text() == '<OBJECT-NAME>Earth'
        
        
        

if __name__ in '__main__':
    pytest.main(args=[__file__])


"""
Handling of PSG's Global Emission Spectra (GlobES) application
"""
from typing import Any, Tuple, List
import numpy as np
from astropy import units as u, constants as c

from . import structure
from ..cfg.models import EquilibriumAtmosphere
from ..cfg.base import Molecule, Aerosol, Profile
from ..settings import atmosphere_type_dict as mtype, aerosol_type_dict as atype, get_setting
from .decoder import GCMdecoder, sep_header

ANGLE_UNIT = u.deg
DTYPE = np.float32


class GCM:
    """
    Global Circulation Model (GCM)

    Parameters
    ----------
    header : str
        The header of the GCM. This tells PSG how to interpret the data.
    dat : np.ndarray
        The data of the GCM.
    """
    KEY = 'ATMOSPHERE-GCM-PARAMETERS'
    BIN_KEY = 'BINARY'
    ENCODING = 'UTF-8'

    def __init__(
        self,
        header: str,
        dat: np.ndarray
    ):
        self.header = header
        self.dat = dat

    @classmethod
    def from_cfg(cls, d: dict):
        """
        Read a GCM from a config dict.

        Parameters
        ----------
        d : dict
            A dictionary read from a PSG config file.
        """
        if cls.KEY not in d:
            return None
        if cls.BIN_KEY not in d:
            return None
        header = d[cls.KEY]
        bin_dat: bytes = d[cls.BIN_KEY]
        dat = np.frombuffer(bin_dat, dtype=np.float32)
        return cls(header, dat)

    @property
    def content(self) -> bytes:
        """
        Get the content of the GCM in a format that PSG can read.

        Returns
        -------
        bytes
            The content of the GCM.
        """
        params_line = bytes(
            f'<{self.KEY}>{self.header}\n', encoding=self.ENCODING)
        start_tag = bytes(f'<{self.BIN_KEY}>', encoding=self.ENCODING)
        end_tag = bytes(f'</{self.BIN_KEY}>', encoding=self.ENCODING)
        dat = self.dat.tobytes(order='C')
        return params_line + b'\n' + start_tag + dat + end_tag


class PyGCM:
    """
    A Python container for GCM data.

    Parameters
    ----------
    pressure : pypsg.globes.structure.Pressure
        The pressure variable.
    temperature : pypsg.globes.structure.Temperature
        The temperature variable.
    wind_u : pypsg.globes.structure.Wind, optional
        The west-to-east wind variable.
    wind_v : pypsg.globes.structure.Wind, optional
        The south-to-north wind variable.
    tsurf : pypsg.globes.structure.SurfaceTemperature, optional
        The surface temperature variable.
    psurf : pypsg.globes.structure.SurfacePressure, optional
        The surface pressure variable.
    albedo : pypsg.globes.structure.Albedo, optional
        The albedo variable.
    emissivity : pypsg.globes.structure.Emissivity, optional
        The emissivity variable.
    *args : pypsg.globes.structure.Molecule | pypsg.globes.structure.Aerosol | pypsg.globes.structure.AerosolSize | pypsg.globes.structure.Surface
        Additional variables.
    """
    _key_order = [
        'wind_u',
        'wind_v',
        'pressure',
        'temperature',
        'tsurf',
        'psurf',
        'albedo',
        'emissivity'
    ]

    def __init__(
        self,
        pressure: structure.Pressure,
        temperature: structure.Temperature,
        *args: structure.Molecule | structure.Aerosol | structure.AerosolSize | structure.Surface,
        wind_u: structure.Wind = None,
        wind_v: structure.Wind = None,
        tsurf: structure.SurfaceTemperature = None,
        psurf: structure.SurfacePressure = None,
        albedo: structure.Albedo = None,
        emissivity: structure.Emissivity = None,
        lon_start: float = -180.,
        lat_start: float = -90.,
        desc: str = None,
    ):
        self.pressure = pressure
        self.temperature = temperature
        self.wind_u = structure.Wind.zero(
            'wind_u', pressure.shape) if wind_u is None else wind_u
        self.wind_v = structure.Wind.zero(
            'wind_v', pressure.shape) if wind_v is None else wind_v
        self.tsurf = tsurf
        self.psurf = psurf
        self.albedo = albedo
        self.emissivity = emissivity
        for arg in args:
            self.__setattr__(arg.name, arg)
        self.lon_start = lon_start
        self.lat_start = lat_start
        self.desc = desc
        self.variables = self._variables

    def __setattr__(self, __name: str, __value: Any) -> None:
        """
        Enforce shapes match pressure data.
        """
        if __name == 'pressure':  # must be set first
            pass
        else:
            nlayers, nlon, nlat = self.shape
            if __value is None:
                pass
            elif isinstance(__value, structure.Variable3D):
                if __value.shape != (nlayers, nlon, nlat):
                    raise ValueError(
                        f'Dimension mismatch: {__value.shape} != ({nlayers},{nlon},{nlat})')
            elif isinstance(__value, structure.Variable2D):
                if __value.shape != (nlon, nlat):
                    raise ValueError(
                        f'Dimension mismatch: {__value.shape} != ({nlon},{nlat})')

        super().__setattr__(__name, __value)

    @property
    def shape(self) -> Tuple[int, int, int]:
        """
        Get the shape of the data.

        Returns
        -------
        nlayers : int
            The number of layers.
        nlon : int
            The number of longitudes.
        nlat : int
            The number of latitudes.
        """
        nlayers, nlon, lat = self.pressure.shape
        return nlayers, nlon, lat

    @property
    def _variables(self) -> List[structure.Variable]:
        """
        Enforce variable order is consistent across outputs.
        """
        variables = []
        for key in self._key_order:
            value = self.__getattribute__(key)
            if value is not None:
                variables.append(value)
        for key, value in self.__dict__.items():
            if key not in self._key_order and isinstance(value, structure.Variable):
                variables.append(value)
        return variables

    @property
    def dlon(self) -> u.Quantity:
        """
        Returns the longitudinal grid spacing.

        Returns
        -------
        dlon : u.Quantity
            The longitudinal grid spacing.
        """

        _, nlon, _ = self.shape
        return 360*u.deg / (nlon)

    @property
    def dlat(self) -> u.Quantity:
        """
        Returns the latitudinal grid spacing.

        Returns
        -------
        dlat : u.Quantity
            The latitudinal grid spacing.
        """

        _, _, nlat = self.shape
        return 180*u.deg / (nlat)

    @property
    def lons(self) -> np.ndarray:
        """
        Get the longitude values.
        """
        _, nlon, _ = self.shape
        return self.lon_start + np.arange(nlon+1) * self.dlon.to_value(ANGLE_UNIT)

    @property
    def lats(self) -> np.ndarray:
        """
        Get the latitude values.
        """
        _, _, nlat = self.shape
        return self.lat_start + np.arange(nlat+1) * self.dlat.to_value(ANGLE_UNIT)

    @property
    def header(self):
        """
        Constructs the string value of the PSG ``<ATMOSPHERE-GCM-PROPERTIES>`` parameter.

        Returns
        -------
        header : str
            The header of the GCM.
        """
        nlayer, nlon, nlat = self.shape
        coords = f'{nlon},{nlat},{nlayer},{self.lon_start:.1f},{self.lat_start:.1f},{self.dlon.to_value(ANGLE_UNIT):.2f},{self.dlat.to_value(ANGLE_UNIT):.2f}'
        variables = self.variables[2:]  # Skip both winds.
        var_names = ['Winds'] + [v.name for v in variables]
        return f'{coords},{",".join(var_names)}'

    @property
    def flat(self) -> np.ndarray:
        """
        A flat array with all the data layed out as expected.

        Returns
        -------
        np.ndarray
            The flattened array.
        """
        return np.concatenate([v.flat.astype(DTYPE) for v in self.variables]).astype(DTYPE)

    @property
    def molecules(self):
        """
        Get the molecules.
        """
        return [molecule for molecule in self.__dict__.values() if isinstance(molecule, structure.Molecule)]

    @property
    def aerosols(self):
        """
        Get the aerosols.
        """
        return [aerosol for aerosol in self.__dict__.values() if isinstance(aerosol, structure.Aerosol)]

    @property
    def aerosol_sizes(self):
        """
        Get the aerosol sizes.
        """
        return [aerosol_size for aerosol_size in self.__dict__.values() if isinstance(aerosol_size, structure.AerosolSize)]

    def update_params(self, atmosphere: EquilibriumAtmosphere = None):
        """
        Update the config.
        """
        if not isinstance(atmosphere, EquilibriumAtmosphere):
            atmosphere = EquilibriumAtmosphere()

        gases = [molec.name for molec in self.molecules]
        aeros = [aerosol.name for aerosol in self.aerosols]
        gas_types = [f'HIT[{mtype[gas]}]' if isinstance(
            mtype[gas], int) else mtype[gas] for gas in gases]
        aerosol_types = [atype[aerosol] for aerosol in aeros]

        molecules = [Molecule(gas, gas_type, 1)
                     for gas, gas_type in zip(gases, gas_types)]
        aerosols = [Aerosol(aerosol, aerosol_type, 1, 1)
                    for aerosol, aerosol_type in zip(aeros, aerosol_types)]
        
        pressure_dat = self.pressure.dat[:, 0, 0].to_value(self.pressure.psg_unit)
        temperature_dat = self.temperature.dat[:, 0, 0].to_value(self.temperature.psg_unit)
        pressure = Profile('Pressure', pressure_dat, u.bar)
        temperature = Profile('Temperature', temperature_dat, u.K)
        

        atmosphere.description = self.desc
        atmosphere.molecules = tuple(molecules) if len(molecules) > 0 else None
        atmosphere.aerosols = tuple(aerosols) if len(aerosols) > 0 else None
        atmosphere.profile = tuple([pressure, temperature])
        return atmosphere

    @classmethod
    def from_decoder(cls, decoder: GCMdecoder):
        """
        Read a GCM from a decoder.
        """
        coords, variables = sep_header(decoder.header)
        pressure = decoder['Pressure']
        pressure = structure.Pressure(10**pressure * u.bar)
        temperature = decoder['Temperature']
        temperature = structure.Temperature(temperature * u.K)            
        
        args = {}
        
        molecules = decoder.get_molecules()
        for molecule in molecules:
            args[molecule] = structure.Molecule(molecule, (10**decoder[molecule])*u.dimensionless_unscaled)
        aerosols, aerosol_sizes = decoder.get_aerosols()
        for aerosol, aerosol_size in zip(aerosols, aerosol_sizes):
            args[aerosol] = structure.Aerosol(aerosol, (10**decoder[aerosol])*u.dimensionless_unscaled)
            args[aerosol_size] = structure.AerosolSize(
                aerosol_size, 10**(decoder[aerosol_size])*u.m)
        
        kwargs = {}
        if 'Wind' in variables:
            wind = decoder['Wind']
            wind_u = wind[0, :, :, :] * u.m / u.s
            wind_v = wind[1, :, :, :] * u.m / u.s
            kwargs['wind_u'] = structure.Wind('wind_u', wind_u)
            kwargs['wind_v'] = structure.Wind('wind_v', wind_v)
        if 'Tsurf' in variables:
            tsurf = decoder['Tsurf']
            kwargs['tsurf'] = structure.SurfaceTemperature(tsurf * u.K)
        if 'Psurf' in variables:
            psurf = decoder['Psurf']
            kwargs['psurf'] = structure.SurfacePressure(10**psurf * u.bar)
        if 'Albedo' in variables:
            albedo = decoder['Albedo']
            kwargs['albedo'] = structure.Albedo(albedo*u.dimensionless_unscaled)
        if 'Emissivity' in variables:
            emissivity = decoder['Emissivity']
            kwargs['emissivity'] = structure.Emissivity(emissivity*u.dimensionless_unscaled)

        _, _, _, lon_start, lat_start, _, _ = coords
        kwargs['lon_start'] = float(lon_start)
        kwargs['lat_start'] = float(lat_start)

        return cls(pressure,temperature, *args.values(), **kwargs)

    @classmethod
    def from_bytes(cls, header: str, binary: bytes):
        """
        Read a GCM from bytes.

        Parameters
        ----------
        header : str
            The header of the GCM.
        binary : bytes
            The binary data of the GCM.
        """
        decoder = GCMdecoder(header, np.frombuffer(binary,dtype='float32'))
        return cls.from_decoder(decoder)

    @classmethod
    def from_cfg(cls, d: dict):
        """
        Read a GCM from a config dict.

        Parameters
        ----------
        d : dict
            A dictionary read from a PSG config file.
        """
        header = d['ATMOSPHERE-GCM-PARAMETERS']
        dat = d['BINARY']
        return cls.from_bytes(header, dat)

    @property
    def content(self)-> bytes:
        """
        Get the content of the GCM in a format that PSG can read.

        Returns
        -------
        bytes
            The content of the GCM.
        """
        gcm_params = b'<ATMOSPHERE-GCM-PARAMETERS>' + self.header.encode(get_setting('encoding'))
        binary = b'<BINARY>' + self.flat.tobytes(order='C') + b'</BINARY>'
        return gcm_params + b'\n' + binary
    
    def altitude(self, mass: u.Quantity, radius: u.Quantity, mean_molecular_mass: float) -> u.Quantity:
        """
        Use the equation of hydrostatic equilibrium to calculate the
        altitude of each grid point.

        Parameters
        ----------
        mass : u.Quantity
            The mass of the planet.
        radius : u.Quantity
            The radius of the planet's surface (the bottom layer).
        mean_molecular_mass : float
            The mean molecular mass of the atmosphere.

        Returns
        -------
        z : u.Quantity
            The altitude of each grid point.
        """
        pressure = self.pressure
        temperature = self.temperature
        nlayer, nlon, nlat = self.shape
        z_unit = u.km
        z = np.zeros(shape=(nlayer, nlon, nlat))
        for i in range(nlayer-1):
            pressure_bottom = pressure.dat[i, :, :].to(u.bar)
            pressure_top = pressure.dat[i+1, :, :].to(u.bar)
            dP = pressure_top - pressure_bottom
            # pylint: disable-next=no-member
            rho = mean_molecular_mass*u.Unit('g mol-1') * \
                (pressure_bottom + 0.5*dP) / c.R / temperature.dat[i, :, :]
            distance_from_planet_center = radius + z[i, :, :]*z_unit
            # pylint: disable-next=no-member
            accel_due_to_gravity = c.G * mass / distance_from_planet_center**2
            dz = -dP / rho / accel_due_to_gravity
            z[i+1, :, :] = z[i, :, :] + dz.to_value(z_unit)
        return z*z_unit

    @staticmethod
    def _column_gas(
        molecule: structure.Molecule,
        pressure: u.Quantity,
        temperature: u.Quantity,
        altitude: u.Quantity
    ) -> u.Quantity:
        """
        Get the column density of a gas.

        Parameters
        ----------
        molecule : structure.Molecule
            The molecule to get the column density of.
        pressure : u.Quantity
            The pressure at each grid point.
        temperature : u.Quantity
            The temperature at each grid point.
        altitude : u.Quantity
            The altitude at each grid point.

        Returns
        -------
        u.Quantity
            The column density of the gas.
        """
        abundance = molecule.dat.to(u.dimensionless_unscaled)
        partial_pressure = pressure * abundance
        heights = np.diff(altitude, axis=0)
        # pylint: disable-next=no-member
        surface_density: u.Quantity = partial_pressure[:-
                                                       1] * heights / c.R / temperature[:-1]
        return surface_density.sum(axis=0).to(u.mol/u.cm**2)

    @staticmethod
    def _column_aerosol(
        aerosol: structure.Aerosol,
        pressure: u.Quantity,
        temperature: u.Quantity,
        altitude: u.Quantity,
        mean_molecular_mass: float
    ):
        """
        Get the column density of an aerosol.

        Parameters
        ----------
        aerosol : structure.Aerosol
            The aerosol to get the column density of.
        pressure : u.Quantity
            The pressure at each grid point.
        temperature : u.Quantity
            The temperature at each grid point.
        altitude : u.Quantity
            The altitude at each grid point.
        mean_molecular_mass : float
            The mean molecular mass of the atmosphere.

        Returns
        -------
        u.Quantity
            The column density of the aerosol.
        """
        mass_frac = aerosol.dat.to_value(u.dimensionless_unscaled)[:-1,:,:]
        heights = np.diff(altitude, axis=0)
        mean_molecular_mass = mean_molecular_mass * u.g/u.mol
        # pylint: disable-next=no-member
        moles_of_gas_per_cm2 = pressure[:-1]*heights / c.R / temperature[:-1]
        mass_of_gas_per_cm2 = moles_of_gas_per_cm2 * mean_molecular_mass
        mass_of_aero_per_cm2 = mass_of_gas_per_cm2 * mass_frac
        return mass_of_aero_per_cm2.sum(axis=0).to(u.kg/u.cm**2)

    def column(
        self,
        var: str,
        mass: u.Quantity,
        radius: u.Quantity,
        mean_molecular_mass: float
    ):
        """
        Get a 2D column density map of a GCM variable.

        Parameters
        ----------
        var : str
            The name of the variable to get the column density of.
        mass : u.Quantity
            The mass of the planet.
        radius : u.Quantity
            The radius of the planet.
        mean_molecular_mass : float
            The mean molecular mass of the atmosphere.

        Returns
        -------
        u.Quantity
            The column density of the variable. The unit depends on the type of variable.
        """
        variable = self.__getattribute__(var)
        if not isinstance(variable, structure.Variable3D):
            raise ValueError(f'{var} is not a 3D variable.')

        pressure = self.pressure.dat.to(u.bar)
        altitude = self.altitude(mass, radius, mean_molecular_mass)
        temperature = self.temperature.dat.to(u.K)
        if isinstance(variable, structure.Molecule):
            return self._column_gas(variable, pressure, temperature, altitude)
        elif isinstance(variable, structure.Aerosol):
            return self._column_aerosol(variable, pressure, temperature, altitude, mean_molecular_mass)
        else:
            raise ValueError(f'{var} is not a molecule or aerosol.')



"""
CFG
===

Dealing with PSG config files.
"""

from pypsg.cfg.config import PyConfig, BinConfig, ConfigTooLongWarning
from pypsg.cfg.models import (
    Target,
    Geometry,
    Atmosphere,
    NoAtmosphere,
    EquilibriumAtmosphere,
    ComaAtmosphere,
    Surface,
    Generator,
    Telescope,
    SingleTelescope,
    Interferometer,
    Coronagraph,
    AOTF,
    LIDAR,
    Noise,
    Noiseless,
    RecieverTemperatureNoise,
    ConstantNoise,
    ConstantNoiseWithBackground,
    PowerEquivalentNoise,
    Detectability,
    CCD
)


"""
PSG Exceptions
"""

class PSGError(Exception):
    """
    The base class for all PSG exceptions.
    """

class GlobESError(PSGError):
    """
    Error in GlobES Application
    """

class PUMASError(PSGError):
    """
    Error in PUMAS Application
    """

class UnknownPSGError(PSGError):
    """
    Unknown PSG Error. If encountered in production submit an issue.
    """

class PSGConnectionError(PSGError):
    """
    PSG Connection Error
    """

class PSGMultiError(PSGError):
    """
    Multiple PSG Errors.
    """

class PSGWarning(UserWarning):
    """
    The base class for all PSG warnings.
    """

class PUMASWarning(PSGWarning):
    """
    Warning in PUMAS Application
    """

class GeneratorWarning(PSGWarning):
    """
    Warning in Generator Application
    """

class UnknownPSGWarning(PSGWarning):
    """
    Unknown PSG Warning. If encountered in production submit an issue.
    """


"""
VSPEC gcm decoder module
"""

from pathlib import Path
import json
import numpy as np
from astropy import units as u, constants as c
from typing import Tuple, List

MOLEC_DATA_PATH = Path(__file__).parent / 'molec.json'


def get_gcm_binary(config: str or Path or bytes):
    """
    Separate a GCM into it's text and binary components.

    Parameters
    ----------
    config : str, pathlib.Path, bytes
        The filename to read or the bytes to parse

    Returns
    -------
    head : str
        The value of the 'ATMOSPHERE-GCM-PARAMETERS' option.
    dat : np.ndarray
        The data between '<BINARY></BINARY>' tags.

    """
    key = '<ATMOSPHERE-GCM-PARAMETERS>'
    start = b'<BINARY>'
    end = b'</BINARY>'
    if isinstance(config, (str, Path)):
        with open(config, 'rb') as file:
            fdat = file.read()
    else:
        fdat = config
    header, dat = fdat.split(start)
    dat = dat.replace(end, b'')
    dat = np.frombuffer(dat, dtype='float32')
    for line in str(header).split(r'\n'):
        if key in line:
            return line.replace(key, ''), np.array(dat)


def sep_header(header):
    fields = header.split(',')
    coords = fields[:7]
    var = fields[7:]
    return coords, var


class GCMdecoder:
    """
    A container for a PSG GCM binary.
    
    This class can read a PSG config containing a GCM and parse the bytes sequence.
    
    Parameters
    ----------
    header : str
        The header of the GCM.
    dat : bytes
        The binary data of the GCM.
    """
    
    DOUBLE = ['Winds']
    """
    Keywords with two 3D arrays.
    """
    FLAT = ['Tsurf', 'Psurf', 'Albedo', 'Emissivity']
    """
    Keywords corresponding to 2D arrays.
    """

    def __init__(self, header:str, dat:bytes):
        self.header = header
        self.dat = dat

    @classmethod
    def from_psg(cls, config: str or Path or bytes):
        """
        Construct a ``GCMdecoder`` from PSG.

        Parameters
        ----------
        config : str, pathlib.Path, bytes
            The filename to read or the bytes to parse
        """
        head, dat = get_gcm_binary(config)
        return cls(head, dat)

    def rename_var(self, oldname, newname):
        """
        Rename a variable in the GCM.
        
        Parameters
        ----------
        oldname : str
            The old name of the variable.
        newname : str
            The new name of the variable.
        """
        coords, vars = sep_header(self.header)
        if oldname in vars:
            vars = [newname if var == oldname else var for var in vars]
        else:
            raise KeyError(f'Variable {oldname} not in header.')
        new_header = ','.join(coords+vars)
        self.header = new_header

    def get_shape(self)->Tuple[int, int, int]:
        """
        The shape of the GCM.
        
        Returns
        -------
        tuple
            The shape of the GCM (Nlon,Nlat,Nlayer).
        """
        coord, _ = sep_header(self.header)
        Nlon, Nlat, Nlayer, _, _, _, _ = coord
        return int(Nlon), int(Nlat), int(Nlayer)

    def get_3d_size(self)->int:
        """
        The size of a 3D variable.
        """
        Nlon, Nlat, Nlayer = self.get_shape()
        return Nlon*Nlat*Nlayer

    def get_2d_size(self)->int:
        """
        The size of a 2D variable.
        """
        Nlon, Nlat, _ = self.get_shape()
        return Nlon*Nlat

    def get_lats(self)->np.ndarray:
        """
        Get the latitude points.
        """
        coord, _ = sep_header(self.header)
        _, Nlat, _, _, lat0, _, dlat = coord
        return np.arange(int(Nlat))*float(dlat) + float(lat0)

    def get_lons(self)->np.ndarray:
        """
        Get the longitude points.
        """
        coord, _ = sep_header(self.header)
        Nlon, _, _, lon0, _, dlon, _ = coord
        return np.arange(int(Nlon))*float(dlon) + float(lon0)

    def get_molecules(self)->List[str]:
        """
        Get the names of the molecules in the GCM
        """
        with open(MOLEC_DATA_PATH, 'rt', encoding='UTF-8') as file:
            molec_data = json.loads(file.read())
        _, variables = sep_header(self.header)
        molecs = [var for var in variables if var in molec_data.keys()]
        return molecs

    def get_aerosols(self)->Tuple[List[str], List[str]]:
        """
        Get the names of the aerosols and the aerosol sizes.
        """
        _, variables = sep_header(self.header)
        aerosols = [var for var in variables if var+'_size' in variables]
        aerosol_sizes = [aero+'_size' for aero in aerosols]
        return aerosols, aerosol_sizes

    def __getitem__(self, item):
        _, variables = sep_header(self.header)
        if not item in variables:
            raise KeyError(
                f'{item} not found. Acceptable keys are {variables}')
        else:
            start = 0

            def get_array_length(var):
                if var in self.DOUBLE:
                    return 2*self.get_3d_size(), 'double'
                elif var in self.FLAT:
                    return self.get_2d_size(), 'flat'
                else:
                    return self.get_3d_size(), 'single'

            def package_array(dat, key):
                if key == 'single':
                    Nlon, Nlat, Nlayer = self.get_shape()
                    return dat.reshape(Nlayer, Nlon, Nlat)
                elif key == 'flat':
                    Nlon, Nlat, Nlayer = self.get_shape()
                    return dat.reshape(Nlon, Nlat)
                elif key == 'double':
                    Nlon, Nlat, Nlayer = self.get_shape()
                    return dat.reshape(2, Nlayer, Nlon, Nlat)
                else:
                    raise ValueError(f'Unknown value {key}')
            for var in variables:
                size, key = get_array_length(var)
                if item == var:
                    dat = self.dat[start:start+size]
                    return package_array(dat, key)
                else:
                    start += size

    def __setitem__(self, item: str, new_value: np.ndarray):
        """
        set an array
        """
        old_value = self.__getitem__(item)
        if not old_value.shape == new_value.shape:
            raise ValueError('New shape must match old shape.')
        new_value = new_value.astype(old_value.dtype)
        _, variables = sep_header(self.header)

        def get_array_length(var):
            if var in self.DOUBLE:
                return 2*self.get_3d_size(), 'double'
            elif var in self.FLAT:
                return self.get_2d_size(), 'flat'
            else:
                return self.get_3d_size(), 'single'
        start = 0
        for var in variables:
            size, _ = get_array_length(var)
            if item == var:
                dat = new_value.flatten(order='C')
                self.dat[start:start+size] = dat
                return None
            else:
                start += size

    def remove(self, item):
        """
        Remove an item from the gcm
        
        Parameters
        ----------
        item : str
            The name of the item to remove.
        """
        coords, variables = sep_header(self.header)
        if item not in variables:
            return ValueError(f'Unknown {item}')

        def get_array_length(var):
            if var in self.DOUBLE:
                return 2*self.get_3d_size(), 'double'
            elif var in self.FLAT:
                return self.get_2d_size(), 'flat'
            else:
                return self.get_3d_size(), 'single'
        start = 0
        for var in variables:
            size, _ = get_array_length(var)
            if item == var:
                s = slice(start, start+size)
                self.dat = np.delete(self.dat, s)
            else:
                start += size
        new_variables = [var for var in variables if item != var]
        self.header = ','.join(coords+new_variables)

    def copy_config(self, path_to_copy: Path, path_to_write: Path, NMAX=2, LMAX=2, mean_mass=28):
        """
        Copy a PSG config file but overwrite all GCM parameters and data
        
        Parameters
        ----------
        path_to_copy : pathlib.Path
            The path to the config file to copy.
        path_to_write : pathlib.Path
            The path to write the new config file.
        NMAX : int
            PSG NMAX parameter
        LMAX : int
            PSG LMAX parameter
        mean_mass : float
            The mean molecular mass of the atmosphere.
        """
        def replace_line(line):
            if b'<ATMOSPHERE-GCM-PARAMETERS>' in line:
                return bytes('<ATMOSPHERE-GCM-PARAMETERS>' + self.header + '\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-LAYERS>' in line:
                _, _, Nlayer = self.get_shape()
                return bytes(f'<ATMOSPHERE-LAYERS>{Nlayer}\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-NGAS>' in line:
                n_molecs = len(self.get_molecules())
                return bytes(f'<ATMOSPHERE-NGAS>{n_molecs}\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-GAS>' in line:
                molecs = ','.join(self.get_molecules())
                return bytes(f'<ATMOSPHERE-GAS>{molecs}\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-TYPE>' in line:
                with open(MOLEC_DATA_PATH, 'rt', encoding='UTF-8') as file:
                    molec_data = json.loads(file.read())
                molecs = self.get_molecules()
                atm_types = ','.join(
                    [f'HIT[{molec_data[mol]["ID"]}]' for mol in molecs])
                return bytes(f'<ATMOSPHERE-TYPE>{atm_types}\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-ABUN>' in line:
                n_molecs = len(self.get_molecules())
                return bytes(f'<ATMOSPHERE-ABUN>{",".join(["1"]*n_molecs)}\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-UNIT>' in line:
                n_molecs = len(self.get_molecules())
                return bytes(f'<ATMOSPHERE-UNIT>{",".join(["scl"]*n_molecs)}\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-NAERO>' in line:
                n_aero = len(self.get_aerosols()[0])
                return bytes(f'<ATMOSPHERE-NAERO>{n_aero}\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-AEROS>' in line:
                aeros = ','.join(self.get_aerosols()[0])
                return bytes(f'<ATMOSPHERE-AEROS>{aeros}\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-ATYPE>' in line:
                dat = {
                    'Water': 'AFCRL_Water_HRI',
                    'WaterIce': 'Warren_ice_HRI'
                }
                atypes = ','.join([dat[aero]
                                  for aero in self.get_aerosols()[0]])
                return bytes(f'<ATMOSPHERE-ATYPE>{atypes}\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-AABUN>' in line:
                n_aero = len(self.get_aerosols()[0])
                return bytes(f'<ATMOSPHERE-AABUN>{",".join(["1"]*n_aero)}\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-AUNIT>' in line:
                n_aero = len(self.get_aerosols()[0])
                return bytes(f'<ATMOSPHERE-AUNIT>{",".join(["scl"]*n_aero)}\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-ASIZE>' in line:
                n_aero = len(self.get_aerosols()[0])
                return bytes(f'<ATMOSPHERE-ASIZE>{",".join(["1"]*n_aero)}\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-ASUNI>' in line:
                n_aero = len(self.get_aerosols()[0])
                return bytes(f'<ATMOSPHERE-ASUNI>{",".join(["scl"]*n_aero)}\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-NMAX>' in line:
                return bytes(f'<ATMOSPHERE-NMAX>{NMAX}\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-LMAX>' in line:
                return bytes(f'<ATMOSPHERE-LMAX>{LMAX}\n', encoding='UTF-8')
            elif b'<ATMOSPHERE-WEIGHT>' in line:
                return bytes(f'<ATMOSPHERE-WEIGHT>{mean_mass}\n', encoding='UTF-8')
            else:
                return line + b'\n'

        with open(path_to_copy, 'rb') as infile:
            with open(path_to_write, 'wb') as outfile:
                contents = infile.read()
                t, b = contents.split(b'<BINARY>')
                b = b.replace(b'</BINARY>', b'')
                lines = t.split(b'\n')
                for line in lines:
                    outfile.write(replace_line(line))
                outfile.write(b'<BINARY>')
                outfile.write(np.asarray(self.dat, dtype='float32', order='C'))
                outfile.write(b'</BINARY>')

    def get_mean_molec_mass(self)->np.ndarray:
        """
        Get the mean molecular mass at every point on the GCM
        """
        with open(MOLEC_DATA_PATH, 'rt', encoding='UTF-8') as file:
            molec_data = json.loads(file.read())
        Nlon, Nlat, Nlayer = self.get_shape()
        mean_molec_mass = np.zeros(shape=(Nlayer, Nlat, Nlon))*u.g/u.mol
        for mol, dat in molec_data.items():
            mass = dat['mass']
            try:
                data = 10**self[mol]
                mean_molec_mass += data*mass*u.g/u.mol
            except KeyError:
                pass
        return mean_molec_mass

    def get_alt(self, M: u.Quantity, R: u.Quantity)->u.Quantity:
        """
        Get the altitude of each GCM point.
        
        Parameters
        ----------
        M : astropy.units.Quantity
            The mass of the planet.
        R : astropy.units.Quantity
            The radius of the planet.
        
        Returns
        -------
        z : astropy.units.Quantity
            The altitude of each GCM point.
        """
        P = 10**self['Pressure']*u.bar
        T = self['Temperature']*u.K
        m = self.get_mean_molec_mass()
        Nlon, Nlat, Nlayers = self.get_shape()
        z_unit = u.km
        z = [np.zeros(shape=(Nlat, Nlon))]
        for i in range(Nlayers-1):
            dP = P[i+1, :, :] - P[i, :, :]
            rho = m[i, :, :]*(P[i, :, :] + 0.5*dP)/c.R/T[i, :, :]
            r = z[-1]*z_unit + R
            g = M*c.G/r**2
            dz = -dP/rho/g
            z.append((z[-1]*z_unit+dz).to(z_unit).value)
        return z*z_unit

    def get_column_density(self, mol: str, M: u.Quantity, R: u.Quantity,)->u.Quantity:
        """
        Get the column density of a gas at each point on the gcm.
        """
        abn = 10**self[mol]*u.mol/u.mol
        P = 10**self['Pressure']*u.bar
        T = self['Temperature']*u.K
        partial_pressure = P*abn
        alt = self.get_alt(M, R)
        heights = np.diff(alt, axis=0)
        density: u.Quantity = np.sum(
            partial_pressure[:-1]*heights/c.R/T[:-1], axis=0)
        return density.to(u.mol/u.cm**2)

    def get_column_clouds(self, var: str, M: u.Quantity, R: u.Quantity,):
        """
        Get the column density of a cloud at each point on the gcm.
        """
        mass_frac = 10**self[var]*u.kg/u.kg
        P = 10**self['Pressure']*u.bar
        T = self['Temperature']*u.K
        molar_mass = self.get_mean_molec_mass()
        alt = self.get_alt(M, R)
        heights = np.diff(alt, axis=0)
        gas_mass_density = P[:-1]*heights/c.R/T[:-1]*molar_mass[:-1]  # g cm-2
        mass_density = np.sum(mass_frac[:-1]*gas_mass_density, axis=0).cgs
        return mass_density.to(u.kg/u.cm**2)



"""
PyPSG top-level module
======================
"""
__version__ = '0.2.1'
from .request import APICall, PSGResponse
from . import cfg
from .cfg import PyConfig
from .rad import PyRad
from .lyr import PyLyr
from . import settings
from . import units
from . import docker
from . import globes



"""
Test pypsg.request module.
"""
from pathlib import Path
import pytest
from astropy import units as u

from pypsg import PyConfig, APICall, PyRad, PyLyr
from pypsg import request as psgrequest


@pytest.fixture
def default_cfg():
    """
    Get a simple default configuration object.
    """
    return PyConfig.from_file(Path(__file__).parent / 'data' / 'simple.cfg')


@pytest.fixture
def advanced_cfg():
    """
    A more complex configuration object. 
    """
    return PyConfig.from_file(Path(__file__).parent / 'data' / 'advanced.cfg')
# pylint: disable=redefined-outer-name


def test_api_init(default_cfg):
    """
    Test api intialization.
    """
    api = APICall(default_cfg, 'rad', 'globes', 'testurl')
    assert api.url == 'testurl'
    assert api.app == 'globes'
    assert api.type == 'rad'
    assert api.cfg.target.object.value == 'Exoplanet'


def test_api_call_rad(default_cfg,psg_url):
    """
    Test api call.
    """
    api = APICall(default_cfg, 'rad',url=psg_url)
    response = api()
    assert isinstance(response, psgrequest.PSGResponse)
    assert isinstance(response.rad, PyRad)
    assert isinstance(response.rad.wl, u.Quantity)


def test_api_call_all(default_cfg, psg_url):
    """
    Test api call.
    """
    api = APICall(default_cfg, 'all',url=psg_url)
    response = api()
    assert isinstance(response, psgrequest.PSGResponse)
    assert isinstance(response.cfg, PyConfig)
    assert isinstance(response.rad, PyRad)


def test_api_call_multiple(default_cfg):
    """
    Test api call.
    """
    with pytest.raises(NotImplementedError):
        _ = APICall(default_cfg, ('rad', 'cfg'))


def test_api_call_advanced(advanced_cfg, psg_url):
    """
    Test api call.
    """
    api = APICall(advanced_cfg, 'all',url=psg_url)
    response = api()
    assert isinstance(response, psgrequest.PSGResponse)
    assert isinstance(response.rad, PyRad)
    assert isinstance(response.lyr, PyLyr)
    assert isinstance(response.noi, PyRad)
    assert isinstance(response.cfg, PyConfig)
    assert response.rad.wl.unit == u.micron


if __name__ == '__main__':
    pytest.main(args=[__file__])



import pytest
from astropy import units as u

from pypsg import units

def test_ppmv():
    abn = 100*units.ppm
    assert abn.to_value(u.dimensionless_unscaled) == pytest.approx(1e-4,abs=1e-7)
def test_ppbv():
    abn = 1*units.ppb
    assert abn.to_value(u.dimensionless_unscaled) == pytest.approx(1e-9,abs=1e-12)
def test_pptv():
    abn = 1*units.ppt
    assert abn.to_value(u.dimensionless_unscaled) == pytest.approx(1e-12,abs=1e-12)


"""
VSPEC GCM structure
"""
from abc import ABC
import warnings
from astropy import units as u
from astropy.units.core import Unit
# from astropy.units.photometric import Unit
from astropy.units.quantity import Quantity
# from astropy.units.si import Unit


import numpy as np

DTYPE = 'float32'

class Variable(ABC):
    """
    The base class representing a GCM (General Circulation Model) variable.

    Parameters
    ----------
    name : str
        The name of the variable.
    psg_unit : astropy.units.Unit
        The unit assumed by the PSG GlobES app.
    dat : astropy.units.Quantity
        The data values of the variable as a `astropy.units.Quantity` object.

    Attributes
    ----------
    name : str
        The name of the variable.
    psg_unit : astropy.units.Unit
        The unit assumed by the PSG GlobES app.
    dat : astropy.units.Quantity
        The data values of the variable as a `astropy.units.Quantity` object.

    Methods
    -------
    flat() -> np.ndarray:
        Returns a flattened version of the data values of the variable as a
        NumPy array, with the physical unit converted to the specified `psg_unit`.
        The array is of dtype 'float32' and the flattening order is 'C'. This is the
        format in which PSG assumes GCM binaries are written.

    shape() -> Tuple[int]:
        Returns the shape of the data values of the variable as a tuple of integers.

    """

    def __init__(
        self,
        name: str,
        psg_unit: u.Unit,
        dat: u.Quantity
    ):
        if not isinstance(dat, u.Quantity):
            raise ValueError(f'dat must be a Quantity object, not {type(dat)}.')
        
        self.name = name
        self.psg_unit = psg_unit
        self.dat = dat

    @property
    def flat(self) -> np.ndarray:
        """
        Get a flattened version of the array.

        Returns a flattened version of the data values of the variable as a
        NumPy array, with the physical unit converted to the specified `psg_unit`.
        The array is of dtype 'float32' and the flattening order is 'C'. This is the
        format in which PSG assumes GCM binaries are written.

        Returns
        -------
        np.array
            The flattened array.
        """
        if self.dat.ndim == 1:
            return self.dat.to_value(self.psg_unit).astype(DTYPE).flatten('C')
        if self.dat.ndim == 2:
            axes = (0, 1)
            return np.swapaxes(self.dat.to_value(self.psg_unit).astype(DTYPE), *axes).flatten('C')
            # return self.dat.to_value(self.psg_unit).T[:,::-1].astype(DTYPE).flatten('C')
        else:
            axes = (1, 2)
        return np.swapaxes(self.dat.to_value(self.psg_unit).astype(DTYPE), *axes).flatten('C')

    @property
    def shape(self) -> tuple:
        """
        Get the shape of the data.

        Returns
        -------
        tuple
            The shape of the data.
        """
        return self.dat.shape


class Variable2D(Variable):
    """
    Variable subclass for 2D variables.
    """

    def __init__(self, name: str, psg_unit: Unit, dat: Quantity):
        if not dat.ndim == 2:
            raise ValueError(f'dat must have a shape of 2, not {dat.ndim}.')
        super().__init__(name, psg_unit, dat)


class Variable3D(Variable):
    """
    Variable subclass for 3D variables.
    """

    def __init__(self, name: str, psg_unit: Unit, dat: Quantity):
        if not dat.ndim == 3:
            raise ValueError(f'dat must have a shape of 3, not {dat.ndim}.')
        super().__init__(name, psg_unit, dat)


class Wind(Variable3D):
    """
    A subclass of Variable representing wind of a planet.

    Parameters
    ----------
    name : str
        The name of the wind variable. Usually either 'U' or 'V'.
    dat : astropy.units.Quantity
        The data values of the wind variable as a `astropy.units.Quantity` object.

    Attributes
    ----------
    name : str
        The name of the wind variable.
    psg_unit : astropy.units.Unit
        The physical unit of the wind variable, which is set to 'm s-1' for wind variables.
    dat : astropy.units.Quantity
        The data values of the wind variable as a `astropy.units.Quantity` object.

    """

    def __init__(self, name: str, dat: u.Quantity):
        if name not in ['wind_u', 'wind_v']:
            msg = f'Wind variable name must be either "wind_u" or "wind_v", not {name}. '
            msg += 'This is to prevent confusion by the PyGCM `__setattr__` method, which assigns unknown variables based on name.'
            warnings.warn(msg, RuntimeWarning)
        super().__init__(name, u.Unit('m s-1'), dat)

    @classmethod
    def constant(
        cls,
        name: str,
        value: u.Quantity,
        shape: tuple
    ):
        """
        Creates a Wind object with constant wind values.

        Parameters
        ----------
        name : str
            The name of the wind variable.
        value : astropy.units.Quantity
            The constant value of the wind variable.
        shape : tuple
            The shape of the wind variable data array.

        Returns
        -------
        Wind:
            A Wind object with constant wind values specified by the given `value` and `shape`.
        """

        dat = np.ones(shape=shape)*value
        return cls(name, dat)

    @classmethod
    def zero(cls, name, shape):
        """
        Creates a Wind object with zero wind values.

        Parameters
        ----------
        name : str
            The name of the wind variable.
        shape : tuple
            The shape of the wind variable data array.

        Returns
        -------
        Wind
            A Wind object with zero wind values specified by the given `shape`.
        """

        dat = np.zeros(shape=shape)
        return cls(name, dat*u.Unit('m s-1'))


class Pressure(Variable3D):
    """
    A subclass of `Variable` representing pressure of a planet.

    Parameters
    ----------
    dat : astropy.units.Quantity
        The data values of the pressure variable as a `astropy.units.Quantity` object.

    Attributes
    ----------
    name : str
        The name of the pressure variable, which is set to 'Pressure'.
    psg_unit : astropy.units.Unit
        The physical unit of the pressure variable, which is set to a logarithmic unit of 'bar' using the `astropy.units` module.
    dat : astropy.units.Quantity
        The data values of the pressure variable as a `astropy.units.Quantity` object.


    """

    def __init__(self, dat: u.Quantity):
        super().__init__('Pressure', u.LogUnit(u.bar), dat)
    
    @property
    def flat(self):
        return super().flat
        

    @classmethod
    def from_profile(cls, profile: u.Quantity, shape: tuple):
        """
        Creates a Pressure object with pressure values from a profile.

        Parameters
        ----------
        profile : astropy.units.Quantity
            The profile of pressure values for each layer.
        shape : tuple
            The shape of the pressure variable data array.

        Returns
        -------
        Pressure:
            A Pressure object with pressure values specified by the given `profile` and `shape`.
        """

        n_layers = len(profile)
        shape = (n_layers,) + shape
        dat = np.ones(shape=shape) * profile[:, np.newaxis, np.newaxis]
        return cls(dat)

    @classmethod
    def from_limits(
        cls,
        high: u.Quantity,
        low: u.Quantity,
        shape: tuple
    ):
        """
        Creates a Pressure object with pressure values between specified high and low limits.

        Parameters
        ----------
        high : astropy.units.Quantity
            The pressure at the planet surface
        low : astropy.units.Quantity
            The lowest pressure to include in the atmosphere.
        shape : tuple
            The shape of the pressure variable data array.

        Returns
        -------
        Pressure:
            A `Pressure` object with pressure values between the specified
            `high` and `low` limits, with the given `shape`.
        """

        n_layers = shape[0]
        shape2d = shape[1:]
        log_high = np.log10(high.to_value(u.bar))
        log_low = np.log10(low.to_value(u.bar))
        profile = np.logspace(log_high, log_low, n_layers)*u.bar
        return cls.from_profile(profile, shape2d)

    def get_adiabatic_scalar(self, gamma: float) -> np.ndarray:
        """
        Calculates the adiabatic scalar for the pressure variable using a specified gamma value.

        Parameters
        ----------
        gamma : float
            The value of gamma, representing the ratio of specific heats.

        Returns
        -------
        np.ndarray
            The adiabatic scalar as a function of pressure. Multiplying this
            array by the surface temperature will give the temperature at every
            GCM point.
        """
        p_surf = self.dat[0, :, :]
        relative_pressure = (
            self.dat/p_surf[np.newaxis, :, :]).to_value(u.dimensionless_unscaled)
        power = 1 - (1/gamma)
        return relative_pressure ** power


class SurfacePressure(Variable2D):
    """
    A subclass of Variable representing the surface pressure of a planet.

    Parameters
    ----------
    dat : astropy.units.Quantity
        The surface pressure of the planet at every point.

    Attributes
    ----------
    name : str
        The name of the surface pressure variable, which is set to 'Psurf'.
    psg_unit : astropy.units.Unit
        The physical unit of the surface pressure variable, which is set to a logarithmic unit of 'bar' using the `astropy.units` module.
    dat : astropy.units.Quantity
        The data values of the surface pressure variable as a `astropy.units.Quantity` object.

    """

    def __init__(self, dat: u.Quantity):
        super().__init__('Psurf', u.LogUnit(u.bar), dat)
    
    @property
    def flat(self):
        return super().flat

    @classmethod
    def constant(cls, value: u.Quantity, shape: tuple):
        """
        Creates a SurfacePressure object with constant surface pressure values.

        Parameters
        ----------
        value : astropy.units.Quantity
            The constant value of the surface pressure.
        shape : tuple
            The shape of the surface pressure variable data array.

        Returns
        -------
        SurfacePressure
            A `SurfacePressure` object with constant surface pressure
            values specified by the given `value` and `shape`.
        """

        dat = np.ones(shape=shape)*value
        return cls(dat)

    @classmethod
    def from_pressure(cls, pressure: Pressure):
        """
        Creates a `SurfacePressure` object from a `Pressure` object.

        Parameters
        ----------
        pressure : Pressure
            The `Pressure` object from which to extract the surface pressure value.

        Returns
        -------
        SurfacePressure
            A `SurfacePressure` object with the surface pressure value extracted
            from the provided `Pressure` object.
        """

        dat = pressure.dat[0, :, :]
        return cls(dat)


class SurfaceTemperature(Variable2D):
    """
    A subclass of Variable representing the surface temperature of a planet.

    Parameters
    ----------
    dat : astropy.units.Quantity
        The surface temperature of the planet at every point.

    Attributes
    ----------
    name : str
        The name of the surface temperature variable, which is set to 'Tsurf'.
    psg_unit : astropy.units.Unit
        The physical unit of the surface temperature variable, which is set to Kelvin (K) using the `astropy.units` module.
    dat : astropy.units.Quantity
        The data values of the surface temperature variable as a `astropy.units.Quantity` object.

    """

    def __init__(self, dat: u.Quantity):
        super().__init__('Tsurf', u.K, dat)


class Temperature(Variable3D):
    """
    A subclass of Variable representing the temperature of a planet.

    Parameters
    ----------
    dat : astropy.units.Quantity
        The temperature of the planet at every point.

    Attributes
    ----------
    name : str
        The name of the temperature variable, which is set to 'Temperature'.
    psg_unit : astropy.units.Unit
        The physical unit of the temperature variable, which is set to Kelvin (K) using the `astropy.units` module.
    dat : astropy.units.Quantity
        The data values of the temperature variable as a `astropy.units.Quantity` object.

    """

    def __init__(self, dat: u.Quantity):
        super().__init__('Temperature', u.K, dat)

    @classmethod
    def from_adiabat(
        cls,
        gamma: float,
        tsurf: SurfacePressure,
        pressure: Pressure
    ):
        """
        Creates a Temperature object from an adiabatic profile.

        Parameters
        ----------
        gamma : float
            The adiabatic index.
        tsurf : SurfacePressure
            The surface pressure of the planet.
        pressure : Pressure
            The 3D pressure profile of the planet.

        Returns
        -------
        Temperature
            A Temperature object with the temperature values calculated from the adiabatic profile.
        """

        adiabatic_scale = pressure.get_adiabatic_scalar(gamma)
        dat = adiabatic_scale*tsurf.dat[np.newaxis, :, :]
        return cls(dat)


class Molecule(Variable3D):
    """
    A subclass of Variable representing a specific molecule concentration.

    Parameters
    ----------
    name : str
        The name of the molecule variable.
    dat : astropy.units.Quantity
        The concentration of the molecule at every point.

    Attributes
    ----------
    name : str
        The name of the molecule variable.
    psg_unit : astropy.units.Unit
        The physical unit of the molecule concentration, which is set to a logarithmic unit of 'mol mol-1' using the `astropy.units` module.
    dat : astropy.units.Quantity
        The data values of the molecule concentration as a `astropy.units.Quantity` object.


    """

    def __init__(self, name: str, dat: u.Quantity):
        dat = u.Quantity(dat)
        super().__init__(name, u.LogUnit(u.Unit('mol mol-1')), dat)

    @classmethod
    def constant(cls, name: str, val: u.Quantity, shape: tuple):
        """
        Creates a Molecule object with constant concentration values.

        Parameters
        ----------
        name : str
            The name of the molecule variable.
        val : astropy.units.Quantity
            The constant concentration value of the molecule.
        shape : tuple
            The shape of the molecule concentration data array.

        Returns
        -------
        Molecule
            A Molecule object with constant concentration values specified by the given name, value, and shape.
        """

        dat = val * np.ones(shape=shape)
        return cls(name, dat)


class Aerosol(Variable3D):
    """
    A subclass of Variable representing aerosol concentrations.

    Parameters
    ----------
    name : str
        The name of the aerosol variable.
    dat : astropy.units.Quantity
        The aerosol concentration at every point.

    Attributes
    ----------
    name : str
        The name of the aerosol variable.
    psg_unit : astropy.units.Unit
        The physical unit of the aerosol concentration, which is set to a logarithmic unit of 'kg kg-1' using the `astropy.units` module.
    dat : astropy.units.Quantity
        The data values of the aerosol concentration as a `astropy.units.Quantity` object.

    """

    def __init__(self, name: str, dat: u.Quantity):
        super().__init__(name, u.LogUnit(u.Unit('kg kg-1')), dat)

    @classmethod
    def constant(cls, name: str, val: u.Quantity, shape: tuple):
        """
        Creates an Aerosol object with constant aerosol concentration values.

        Parameters
        ----------
        name : str
            The name of the aerosol variable.
        val : astropy.units.Quantity
            The constant aerosol concentration value.
        shape : tuple
            The shape of the aerosol concentration data array.

        Returns
        -------
        Aerosol
            An Aerosol object with constant aerosol concentration values specified by the given name, value, and shape.
        """

        dat = np.ones(shape=shape) * val
        return cls(name, dat)

    @classmethod
    def boyant_exp(cls, name: str, max_val: u.Quantity, max_pressure: u.Quantity, pressure: Pressure):
        """
        Creates an Aerosol object with a concentration that floats at a certain pressure
        and falls off exponentially with height.

        Parameters
        ----------
        name : str
            The name of the aerosol variable.
        max_val : astropy.units.Quantity
            The maximum aerosol concentration value.
        max_pressure : astropy.units.Quantity
            The maximum pressure threshold above which the aerosol concentration is non-neglegable.
        pressure : Pressure
            The pressure profile of the planet.

        Returns
        -------
        Aerosol
            An Aerosol object with aerosol concentration values that fall off exponentially with height.
        """

        dat = np.where(pressure.dat > max_pressure, 1e-20 *
                       u.Unit('kg kg-1'), max_val * (pressure.dat/max_pressure))
        return cls(name, dat)


class AerosolSize(Variable3D):
    """
    A subclass of Variable representing aerosol particle sizes.

    Parameters
    ----------
    name : str
        The name of the aerosol size variable.
    dat : astropy.units.Quantity
        The aerosol particle sizes at every point.

    Attributes
    ----------
    name : str
        The name of the aerosol size variable.
    psg_unit : astropy.units.Unit
        The physical unit of the aerosol particle sizes, which is set to  Log[meters ('m')] using the `astropy.units` module.
    dat : astropy.units.Quantity
        The data values of the aerosol particle sizes as a `astropy.units.Quantity` object.

    """

    def __init__(self, name: str, dat: u.Quantity):
        super().__init__(name, u.LogUnit(u.m), dat)

    @classmethod
    def constant(cls, name: str, val: u.Quantity, shape: tuple):
        """
        Creates an AerosolSize object with constant aerosol particle sizes.

        Parameters
        ----------
        name : str
            The name of the aerosol size variable.
        val : astropy.units.Quantity
            The constant aerosol particle size value.
        shape : tuple
            The shape of the aerosol particle sizes data array.

        Returns
        -------
        AerosolSize
            An AerosolSize object with constant aerosol particle sizes specified by the given name, value, and shape.
        """

        dat = np.ones(shape=shape) * val
        return cls(name, dat)


class Albedo(Variable2D):
    """
    A subclass of Variable representing albedo values.

    Parameters
    ----------
    dat : astropy.units.Quantity
        The albedo values at every point.

    Attributes
    ----------
    name : str
        The name of the albedo variable, set to 'Albedo'.
    psg_unit : astropy.units.Unit
        The physical unit of the albedo values, which is set to dimensionless using the `astropy.units` module.
    dat : astropy.units.Quantity
        The data values of the albedo variable as a `astropy.units.Quantity` object.

    """

    def __init__(self, dat: u.Quantity):
        dat = u.Quantity(dat)
        super().__init__('Albedo', u.dimensionless_unscaled, dat)

    @classmethod
    def constant(cls, val: u.Quantity, shape: tuple):
        """
        Creates an Albedo object with constant albedo values.

        Parameters
        ----------
        val : astropy.units.Quantity
            The constant albedo value.
        shape : tuple
            The shape of the albedo data array.

        Returns
        -------
        Albedo
            An `Albedo` object with constant albedo values specified by the given value, and shape.
        """

        dat = np.ones(shape=shape) * val
        return cls(dat)


class Emissivity(Variable2D):
    """
    A subclass of Variable representing the emisivity of a planet.

    Parameters
    ----------
    dat : astropy.units.Quantity
        The emisivity values at every point.

    Attributes
    ----------
    name : str
        The name of the variable, set to 'Emissivity'.
    psg_unit : astropy.units.Unit
        The physical unit of the albedo values, which is set to dimensionless using the `astropy.units` module.
    dat : astropy.units.Quantity
        The data values of the albedo variable as a `astropy.units.Quantity` object.

    """

    def __init__(self, dat: u.Quantity):
        super().__init__('Emissivity', u.dimensionless_unscaled, dat)

    @classmethod
    def constant(cls, val: u.Quantity, shape: tuple):
        """
        Creates an Emissivity object with constant emissivity values.

        Parameters
        ----------
        val : astropy.units.Quantity
            The constant albedo value.
        shape : tuple
            The shape of the albedo data array.

        Returns
        -------
        emissivity
            An `Emissivity` object with constant values specified by the given value, and shape.
        """

        dat = np.ones(shape=shape) * val
        return cls(dat)


class Surface(Variable2D):
    """
    A subclass of Variable representing the surface of a planet.

    Parameters
    ----------
    name : str
        The name of the surface variable.
    dat : astropy.units.Quantity
        The emisivity values at every point.

    Attributes
    ----------
    name : str
        The name of the albedo variable.
    psg_unit : astropy.units.Unit
        The physical unit of the surface values, which is set to dimensionless using the `astropy.units` module.
    dat : astropy.units.Quantity
        The data values of the albedo variable as a `astropy.units.Quantity` object.

    """

    def __init__(self, name: str, dat: u.Quantity):
        super().__init__(name, u.dimensionless_unscaled, dat)

    @classmethod
    def constant(cls, name: str, val: u.Quantity, shape: tuple):
        """
        Creates a Surface object with constant values.

        Parameters
        ----------
        val : astropy.units.Quantity
            The constant albedo value.
        shape : tuple
            The shape of the albedo data array.

        Returns
        -------
        dat : astropy.units.Quantity
            A `Surface` object with constant values specified by the given value, and shape.
        """

        dat = np.ones(shape=shape) * val
        return cls(name, dat)



"""
Methods to parse config files.
"""
from typing import Union, Dict, Any
from pathlib import Path

import warnings

from pypsg.cfg import models
from pypsg import settings
from pypsg.globes import PyGCM


class ConfigTooLongWarning(UserWarning):
    """
    The PSG configuration is too long,
    and may stop updating soon.
    """


class BinConfig:
    """
    PSG configuration structure.

    Parameters
    ----------
    content : bytes
        The content of the configuration.

    Attributes
    ----------
    enconding : str
        The encoding of the config. Set to 'UTF-8'.
    content : bytes
        The content of the config.
    has_binary : bool
        True if there is a `<BINARY>` section in the config.
    """
    encoding = 'UTF-8'

    def __init__(self, content: bytes):
        self.content = content

    @classmethod
    def from_file(cls, path: Path):
        """
        Read a config from a file.

        Parameters
        ----------
        path : pathlib.Path
            The path to the file.

        Returns
        -------
        Config
            A config constructed using the provided file.
        """
        # warnings.warn('This method has not been tested.',RuntimeWarning)
        with open(path, 'rb') as file:
            content = file.read()
        return cls(content=content)

    @property
    def has_binary(self) -> bool:
        """
        True if the config contains a binary section

        :type: bool
        """
        # warnings.warn('This method has not been tested.',RuntimeWarning)
        return b'<BINARY>' in self.content

    @property
    def binary(self) -> bytes:
        """
        The binary section of the config.

        :type: bytes
        """
        # warnings.warn('This method has not been tested.',RuntimeWarning)
        if not self.has_binary:
            raise ValueError('This config contains no binary section.')
        return self.content.split(b'<BINARY>')[1].split(b'</BINARY>')[0]

    @property
    def dict(self) -> dict:
        """
        A dictionary with all the keyword, value pairs.

        :type: dict
        """
        # warnings.warn('This method has not been tested.',RuntimeWarning)
        content = self.content
        binary = None
        if self.has_binary:
            content = content.split(b'<BINARY>')[
                0] + content.split(b'</BINARY>')[1]
            binary = self.binary
        content = str(content, encoding=self.encoding)
        n_lines = len(content.split('\n'))
        if n_lines > settings.get_setting('cfg_max_lines'):
            warnings.warn('The config is too long.', ConfigTooLongWarning)
        cfg = {}
        for line in content.split('\n'):
            if not (line.isspace() or len(line) == 0 or line[0] == '#'):
                try:
                    end_of_kwd = line.index('>')+1
                except ValueError as err:
                    raise ValueError(f'Invalid config line: {line}') from err
                kwd = line[:end_of_kwd].replace('<', '').replace('>', '')
                val = line[end_of_kwd:]
                cfg[kwd] = val
        if binary is not None:
            cfg['BINARY'] = binary
        return cfg


class PyConfig:
    """
    A configuration in the form of a python object.

    Parameters
    ----------
    target : models.Target, optional
        Target model, by default None
    geometry : models.Geometry, optional
        Geometry model, by default None
    atmosphere : models.Atmosphere, optional
        Atmosphere model, by default None
    surface : models.Surface, optional
        Surface model, by default None
    generator : models.Generator, optional
        Generator model, by default None
    telescope : models.Telescope, optional
        Telescope model, by default None
    noise : models.Noise, optional
        Noise model, by default None
    gcm : globes.PyGCM, optional
        GCM model, by default None
    """

    def __init__(
        self,
        target: models.Target = None,
        geometry: models.Geometry = None,
        atmosphere: models.Atmosphere = None,
        surface: models.Surface = None,
        generator: models.Generator = None,
        telescope: models.Telescope = None,
        noise: models.Noise = None,
        gcm: PyGCM = None
    ):
        self.target: models.Target = target
        if self.target is None:
            self.target = models.Target()

        self.geometry: models.Geometry = geometry
        if self.geometry is None:
            self.geometry = models.Geometry()

        self.atmosphere: Union[
            models.NoAtmosphere, models.EquilibriumAtmosphere, models.ComaAtmosphere
        ] = atmosphere
        if self.atmosphere is None:
            self.atmosphere = models.Atmosphere()

        self.surface: models.Surface = surface
        if self.surface is None:
            self.surface = models.Surface()

        self.generator: models.Generator = generator
        if self.generator is None:
            self.generator = models.Generator()

        self.telescope: Union[
            models.SingleTelescope, models.Interferometer, models.Coronagraph,
            models.AOTF, models.LIDAR
        ] = telescope
        if self.telescope is None:
            self.telescope = models.Telescope()

        self.noise: Union[
            models.Noiseless, models.RecieverTemperatureNoise,
            models.ConstantNoise, models.ConstantNoiseWithBackground,
            models.PowerEquivalentNoise, models.Detectability, models.CCD
        ] = noise
        if self.noise is None:
            self.noise = models.Noise()

        self.gcm: PyGCM | None = gcm

    @classmethod
    def from_dict(cls, d: Dict[str, Any]):
        """
        Construct a PyConfig from a dictionary.

        Parameters
        ----------
        d : dict
            A dictionary representation of a PSG config file,
            in the format `{'OBJECT-NAME':'Earth'}`. Note that the
            keys must all be captial letters.

        Notes
        -----
        * Missing keys, provided they are all uppercase, are ignored.
        * Some fields require more than one key. In the case that only one
          is provided, it will be ignored.
        * The actual parsing of the dictionary is done by each field.
        """
        for key in d:
            if not key.isupper():
                raise ValueError(f'Invalid config key: {key}')
        has_gcm = 'ATMOSPHERE-GCM-PARAMETERS' in d and 'BINARY' in d
        gcm = PyGCM.from_cfg(d) if has_gcm else None
        atmosphere = models.Atmosphere.from_cfg(d)
        if has_gcm and isinstance(atmosphere, models.EquilibriumAtmosphere):
            atmosphere = gcm.update_params(atmosphere)

        return cls(
            target=models.Target.from_cfg(d),
            geometry=models.Geometry.from_cfg(d),
            atmosphere=atmosphere,
            surface=models.Surface.from_cfg(d),
            generator=models.Generator.from_cfg(d),
            telescope=models.Telescope.from_cfg(d),
            noise=models.Noise.from_cfg(d),
            gcm=gcm
        )

    @classmethod
    def from_binaryconfig(cls, config: BinConfig):
        """
        Construct a PyConfig from a BinConfig object.

        Parameters
        ----------
        config : BinConfig
            A BinConfig object.
        """
        return cls.from_dict(config.dict)

    @classmethod
    def from_bytes(cls, config: bytes):
        """
        Construct a PyConfig from bytes.

        Parameters
        ----------
        config : bytes
            The bytes representation of a config file.
        """
        return cls.from_binaryconfig(BinConfig(config))

    @classmethod
    def from_file(cls, path: Path | str):
        """
        Construct a PyConfig from a file.

        Parameters
        ----------
        path : pathlib.Path | str
            The path to the file.
        """
        return cls.from_binaryconfig(BinConfig.from_file(path))

    @property
    def content(self) -> bytes:
        """
        Get the config content as a bytes string.
        """
        lines = []
        for model in [
            self.target,
            self.geometry,
            self.gcm.update_params(self.atmosphere) if self.gcm is not None else self.atmosphere,
            self.surface,
            self.generator,
            self.telescope,
            self.noise
        ]:
            model: models.Model
            c = model.content
            if c != b'':
                lines.append(c)
        if self.gcm is not None:
            lines.append(self.gcm.content)
        return b'\n'.join(lines)

    def to_file(self, path: Path | str):
        """
        Write the config to a file.

        Parameters
        ----------
        path : pathlib.Path | str
            The path to the file.
        """
        with open(path, 'wb') as f:
            f.write(self.content)
